class Diagnostic < ActiveRecord::Base

  has_many :portfolio_item_diagnostics, class_name: '::Portfolio::ItemDiagnostic',
           foreign_key: :diagnostic_id,
           inverse_of: :diagnostic,
           dependent: :restrict_with_error

  has_many :portfolio_items, class_name: '::Portfolio::Item',
           through: :portfolio_item_diagnostics


  has_many :diagnostic_institution_stocks, class_name: '::DiagnosticInstitutionStock',
           foreign_key: :diagnostic_id,
           inverse_of: :diagnostic,
           dependent: :destroy
  accepts_nested_attributes_for :diagnostic_institution_stocks

  has_many :characteristics, class_name: '::Characteristic',
           foreign_key: :diagnostic_id,
           inverse_of: :diagnostic,
           dependent: :destroy
  accepts_nested_attributes_for :characteristics, allow_destroy: true

  has_many :norm_scales, class_name: '::NormScale',
           foreign_key: :diagnostic_id,
           inverse_of: :diagnostic,
           dependent: :destroy
  accepts_nested_attributes_for :norm_scales, allow_destroy: true

  has_and_belongs_to_many :portfolio_specializations, autosave: true

  has_many :diagnostic_imgs, class_name: '::DiagnosticImg',
           foreign_key: :diagnostic_id,
           inverse_of: :diagnostic,
           autosave: true,
           dependent: :destroy

  has_many :diagnostic_borrows, class_name: '::DiagnosticBorrow',
           foreign_key: :diagnostic_id,
           dependent: :destroy

  belongs_to :institution


  mount_uploader :cover_image, LibraryImgUploader
  after_save do
    cover_image.versions.each do |name,version|
      unless version.file.nil?
        diagnostic_img_attributes = {name: name,width_px: 0, height_px: 0, file_type: version.file.extension.downcase,source_url: version.url, is_local: true}
        existing=diagnostic_imgs.where(name: name).take
        if existing.present?
          existing.update_columns(diagnostic_img_attributes)
        else
          diagnostic_imgs.create!(diagnostic_img_attributes)
        end
      end
    end
  end

  validates_format_of :year, with: /\d{4}/
  validates_presence_of :title, :age_from, :age_to
  validates_inclusion_of :min_level, :max_level, in: [0,1,2,3]
  validates_numericality_of :age_from, :age_to
  validates_inclusion_of :is_range_okologie, :is_range_integration, :is_range_intensitat, in: [true,false]
  validates_length_of :description, minimum: 50
  validates_length_of :abbreviation, maximum: 16
  validate { errors.add(:min_level, 'Förderstufenbereich nicht richtig angegeben') unless min_level.to_i <= max_level.to_i }
  validate { errors.add(:age_from, 'Altersbereich nicht richtig angegeben') unless age_from.to_i <= age_to.to_i }
  validate { errors.add(:is_range_okologie, 'Mindestens einen Bereich angeben') if !(is_range_okologie? || is_range_intensitat? || is_range_integration?) }

  after_initialize  unless: :persisted? do
    self.min_level ||= 0
    self.max_level ||= 3
    self.age_from ||= 1
    self.age_to ||= 99
    self.no_image_url ||= '/files/no-diagnostic-image.jpeg'
  end

  before_validation do
    if self.catalog_id.blank? && self.title.present?
      require 'digest/md5'
      string = Digest::MD5.hexdigest(self.title.gsub(/\s+/,'').downcase)
      map = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'.split(//);
      part = string[0..12].to_i(16)
      hash = ''
      while(part != 0)
        hash = String(map[part % map.length ] ) + hash;
        part = part / map.length;
      end
      self.catalog_id = "vsb-#{hash}";
    end
  end




  def short_title
    return abbreviation if title.length > 60 && abbreviation.present?
    return title[0..60] + " ..." if title.length > 60 && abbreviation.blank?
    return (abbreviation + ' ' + title) if title.length <= 60 && abbreviation.present?
    return title if title.length <= 60 && abbreviation.blank?
  end

  def thumbnail
    diagnostic_imgs.to_a.keep_if{|i| i.name=='thumb' }.first
  end

  def ranges
    r = []
    r << 'Integration' if is_range_integration
    r << 'Ökologie' if is_range_okologie
    r << 'Intensität' if is_range_intensitat
    r
  end

  def can_be_deleted?
    ! portfolio_item_diagnostics.any?
  end

  ######################################################################################################################
  ### Suggestions
  ######################################################################################################################
  # Schüler sollen Vorschläge für Literatur bekommen.4
  # TODO: this is apparently a shared interface between diagnostic, advancement_plan and online_diagnostic.


  # get the matching clients for a specific advancement plan
  def suggested_clients
    suggestions = Client
    # has stock in client institution
    suggestions = suggestions.where(institution_id: diagnostic_institution_stocks.is_available.pluck(:institution_id))
    # client is in age range
    if age_to
      suggestions = suggestions.where('bday >= ?', (Date.today - age_to.to_i))
    end
    if age_from
      suggestions = suggestions.where('bday <= ?', (Date.today - age_from.to_i ))
    end
    # client is in level range
    suggestions = suggestions.where('portfolio_level >= ? AND portfolio_level <= ?', min_level ,max_level)
    suggestions.order(updated_at: :desc) # this is to have the most recently active Clients as a suggestions
  end

end
