class Institution < ActiveRecord::Base

  TYPE_SCHOOL='school'
  TYPE_EDUCATIONAL='educational'
  TYPE_EXTERN='extern'
  TYPE_ORGANISATIONAL='organisational'

  include SystemTaskReceivable

  #--- CW File Uploader
  mount_uploader :logo, LogoImageUploader

  #--- Associations
  belongs_to :institution_category
  belongs_to :address, inverse_of:  :institution

  has_many :institution_documents

  has_many :managers
  #has_many :notifications, dependent: :destroy # refs #192

  has_many :child_institutions, :class_name => 'Institution', :foreign_key => 'parent_id'
  belongs_to :parent_institution, :class_name => 'Institution', :foreign_key => 'parent_id'

  has_many :all_parent_institutions, :class_name => 'InstitutionHierarchy', :foreign_key => 'child_id'
  has_many :all_child_institutions, :class_name => 'InstitutionHierarchy', :foreign_key => 'parent_id'

  has_many :career_items, dependent: :nullify # refs #192

  #--- Associations for Organisational, Educational and school
  has_many :teacher_institution_relations, foreign_key: 'institution_id', dependent: :destroy # refs #192
  has_many :teachers, through: :teacher_institution_relations


  has_many :diagnostic_institution_stocks, inverse_of: :institution, foreign_key: 'institution_id', dependent: :destroy # refs #192
  has_many :advancement_plan_institution_stocks, inverse_of: :institution, foreign_key: 'institution_id', dependent: :destroy # refs #192
  has_many :diagnostic_borrows, inverse_of: :institution, foreign_key: 'institution_id', dependent: :destroy # refs #192
  has_many :advancement_plan_borrows, inverse_of: :institution, foreign_key: 'institution_id' , dependent: :destroy # refs #192




  #--- Associations for Educational and school
  has_many :client_groups, foreign_key: 'institution_id', dependent: :destroy # refs #192
  has_many :current_client_groups,
           -> { distinct.joins(:semester).where("semesters.id = ?", SystemSetting.get_instance.semester_id)},
           :class_name => 'ClientGroup',
           :foreign_key => 'institution_id'
  has_many :current_and_last_client_groups,
           -> { distinct.joins(:semester).where("semesters.id IN (?)", [Semester.last ? Semester.last.id : nil,SystemSetting.get_instance.semester_id])},
           :class_name => 'ClientGroup',
           :foreign_key => 'institution_id'
  has_many :clients, foreign_key: 'institution_id'

  #--- Associations for Extern
  has_many :supporters, inverse_of: :institution, foreign_key: 'institution_id', dependent: :nullify # refs #192
  has_many :supported_clients, :through => :supporters, source: :clients

  has_and_belongs_to_many :educational_courses, class_name: '::EducationalCourse',
      inverse_of: :institutions

  def available_educational_courses_rules
    EducationalCourseRule.where(educational_course_id: educational_course_ids).all
  end

  #--- Scopes
  scope :only_type, ->(type) { (type.present? ? where('institutions.institution_type IN (?)', ( type.is_a?(Array) ? type : [type] )) : all)  }

  scope :schools, -> { where('institutions.institution_type = ?', TYPE_SCHOOL) }
  scope :education_facilities, -> { where('institutions.institution_type = ? OR institutions.institution_type = ?', TYPE_EDUCATIONAL, TYPE_SCHOOL) }
  scope :other_facilities, -> { where('institutions.institution_type = ?',TYPE_EXTERN)  }
  scope :organisation_facilities, -> { where('institutions.institution_type = ?', TYPE_ORGANISATIONAL) }
  scope :root_institutions, -> { where('institutions.is_root = ?',true) }

  accepts_nested_attributes_for :address

  before_save :normalize_phone
  before_validation :set_defaults_for_new, on: :create

  #--- General Validations
  validates :institution_type, presence: true, inclusion: [TYPE_SCHOOL,TYPE_EDUCATIONAL,TYPE_EXTERN,TYPE_ORGANISATIONAL]
  validates :name, presence: true, length: { maximum: 255}
  validates :email, email: true
  validates :institution_category_id, presence: true
  validates :phone, :fax, phony_plausible: {normalized_country_code: 'DE'}
  validates :comment, length: { maximum: 1024}
  validates :description, length: { maximum: 1024}

  #--- Educational and school Validations

  validates :allocation_formula_special_ed_level_1,
            :allocation_formula_special_ed_level_2,
            :allocation_formula_special_ed_level_3,
            numericality: { only_integer: true,  equal_to_or_greater_than: 0 },
            if: Proc.new { |i| i.institution_type==TYPE_EDUCATIONAL || i.institution_type==TYPE_SCHOOL }
  validates :parent_id, presence: true,
            if: Proc.new { |i| i.institution_type==TYPE_EDUCATIONAL || i.institution_type==TYPE_SCHOOL }

  #--- School Validations
  validates :school_no, presence: true, length: { maximum: 255},
            if: Proc.new { |i| i.institution_type==TYPE_SCHOOL }

  #--- Educational and school methods
  def allocation_stats
    if @allocation_stats.nil?
      @allocation_stats = Institutions::AllocationStats.new(self)
    end
    @allocation_stats
  end

  def cannot_be_destroyed?
    return true if (institution_type==TYPE_EDUCATIONAL || institution_type==TYPE_SCHOOL) && clients.any?
    return true if (institution_type==TYPE_ORGANISATIONAL) && all_child_institutions.any?
    return true if (institution_type==TYPE_ORGANISATIONAL) && managers.any?
    return false
  end

  def cannot_be_destroyed_reason
    return "Es gibt noch Schüler an dieser Einrichtung. Alle Schüler müssen die Schule verlassen, bevor Sie zerstört wird." if (institution_type==TYPE_EDUCATIONAL || institution_type==TYPE_SCHOOL) && clients.any?
    return "Es gibt noch untergeordnete Einrichtungen. Die Einrichtung darf keine weiteren untergeordneten Einrichtungen haben." if (institution_type==TYPE_ORGANISATIONAL) && all_child_institutions.any?
    return "Es gibt noch Mitarbeiter an der Einrichtung. Alle Mitarbeiter müssen die Einrichtung verlassen, bevor Sie zerstört wird." if(institution_type==TYPE_ORGANISATIONAL) && managers.any?
  end

  private

  def normalize_phone
    set_phony_normalized_numbers [:fax,:phone], :default_country_code => 'DE'
  end

  def set_defaults_for_new
    if institution_type==TYPE_EDUCATIONAL || institution_type==TYPE_SCHOOL || institution_type==TYPE_EXTERN
      self.is_root=false
    end
    if institution_type==TYPE_ORGANISATIONAL
      self.is_root=true
    end
    #if institution_type==TYPE_SCHOOL
    #  self.allocation_formula= 1 # scheint falsch zu sein. allocation formula bedeutet: 1 volle stelle / <allocation_formula> schüler
    #end
  end

end
