module Portfolio
    class Item < ::ActiveRecord::Base
      self.table_name = 'portfolio_items'
      self.inheritance_column = nil

      include ::AuthorConcern
      include ::SystemTaskReceivable
      include ::Publishable

      ENUM_GUARDIAN_LOCATION_PREFERENCE_CURRENT = 0
      ENUM_GUARDIAN_LOCATION_PREFERENCE_GL = 1
      ENUM_GUARDIAN_LOCATION_PREFERENCE_FZ = 2

      # CHECK LIKE: Portfolio::Item.guardian_location_preferences[item.guardian_location_preference] == Portfolio::Item::ENUM_GUARDIAN_LOCATION_PREFERENCE_CURRENT
      enum guardian_location_preferences: {
          "aktuelle Schule" => ENUM_GUARDIAN_LOCATION_PREFERENCE_CURRENT,
          "andere GL Schule" => ENUM_GUARDIAN_LOCATION_PREFERENCE_GL,
          "Förderzentrum" => ENUM_GUARDIAN_LOCATION_PREFERENCE_FZ
      }

      alias_attribute :subject_involved, :client_involved # client is no longer coupled

      belongs_to :portfolio,
                 foreign_key: :portfolio_id,
                 class_name: '::Portfolio::Portfolio',
                 inverse_of:  :portfolio_items
      delegate :portfolio_subject, to: :portfolio
      delegate :institution, :institution_id, to: :portfolio_subject
      delegate :client_group, :client_group_id, to: :portfolio_subject


      belongs_to :portfolio_step,
                 foreign_key: :portfolio_step_id,
                 class_name: '::Portfolio::Step',
                 inverse_of: nil

      belongs_to :portfolio_specialization,
                 foreign_key: :portfolio_specialization_id,
                 class_name: '::PortfolioSpecialization',
                 inverse_of: :portfolio_items

      belongs_to :portfolio_step_result,
                 foreign_key: :portfolio_step_result_id,
                 class_name: '::Portfolio::StepResult',
                 inverse_of: :portfolio_step

      has_many :diagnostics, # Diagnostiken die auch zu einem Bibliothekseintrag gehören. Haben Ergebnisse
               foreign_key: :portfolio_item_id,
               class_name: '::Portfolio::ItemDiagnostic',
               inverse_of:  :portfolio_item, dependent: :destroy # refs #192
      accepts_nested_attributes_for :diagnostics, :allow_destroy => true, reject_if: :all_blank


      has_many :class_diagnostics, # Verfahren, die eine referenz zu einer online resource haben. Können dokumente hochgeladen werden als Ergebnis.
               foreign_key: :portfolio_item_id,
               class_name: '::Portfolio::ItemClassDiagnostic',
               inverse_of:  :portfolio_item, dependent: :destroy # refs #192
      accepts_nested_attributes_for :class_diagnostics, :allow_destroy => true, reject_if: :all_blank

      has_many :advancement_plans, # Fördermethoden, die auch zu einem Bibliothekseintrag gehören. Haben keine Ergebnisse
               foreign_key: :portfolio_item_id,
               class_name: "::Portfolio::ItemAdvancementPlan",
               inverse_of:  :portfolio_item,
               dependent: :destroy # refs #192
      accepts_nested_attributes_for :advancement_plans, :allow_destroy => true, reject_if: :all_blank


      has_many :portfolio_dbrcs,
               foreign_key: :portfolio_item_id,
               class_name: "::Portfolio::Dbrc",
               inverse_of:  :portfolio_item,
               dependent: :destroy # refs #192
      accepts_nested_attributes_for :portfolio_dbrcs, :allow_destroy => true # reject_if: :all_blank


      has_many :portfolio_documents,
               class_name: '::Portfolio::Document',
               foreign_key: :portfolio_item_id,
               inverse_of: :portfolio_item,
               dependent: :destroy # refs #192
      accepts_nested_attributes_for :portfolio_documents, :allow_destroy => true, reject_if: :all_blank

      has_many :portfolio_item_involved_supporters,
               class_name: '::Portfolio::ItemInvolvedSupporter',
               foreign_key: :portfolio_item_id,
               inverse_of: :portfolio_item,
               dependent: :destroy # refs #192

      has_many :involved_supporters,
               through: :portfolio_item_involved_supporters,
               class_name: '::Supporter',
               source: :supporter, validate: false

      has_many :portfolio_item_involved_teachers,
               class_name:  '::Portfolio::ItemInvolvedTeacher',
               foreign_key: :portfolio_item_id,
               inverse_of: :portfolio_item,
               dependent: :destroy # refs #192

      has_many :involved_teachers,
               through: :portfolio_item_involved_teachers,
               class_name: '::Teacher',
               source: :teacher,
               validate: false

      has_many :portfolio_item_involved_guardians,
               class_name: '::Portfolio::ItemInvolvedGuardian',
               foreign_key: :portfolio_item_id,
               inverse_of: :portfolio_item,
               dependent: :destroy # refs #192

      has_many :involved_guardians,
               through: :portfolio_item_involved_guardians,
               class_name: '::Guardian',
               source: :guardian,
               validate: false

      has_many :portfolio_item_involved_people,
               class_name: '::Portfolio::ItemInvolvedPerson',
               foreign_key: :portfolio_item_id,
               inverse_of:  :portfolio_item,
               dependent: :destroy # refs #192
      accepts_nested_attributes_for :portfolio_item_involved_people, :allow_destroy => true, reject_if: :all_blank

      has_many :measurements, # Wie advancement_plans, aber ohne biliotheksreferenz.
               foreign_key: :portfolio_item_id,
               class_name: '::Portfolio::ItemMeasurement',
               inverse_of:  :portfolio_item,
               dependent: :destroy # refs #192
      accepts_nested_attributes_for :measurements, :allow_destroy => true, reject_if: :all_blank

      has_many :diagnostic_measurements, # Wie diagnostics, aber ohne biliotheksreferenz. Können dokumente als ergebnis hochgeladen werden
               foreign_key: :portfolio_item_id,
               class_name: '::Portfolio::ItemDiagnosticMeasurement',
               inverse_of:  :portfolio_item,
               dependent: :destroy # refs #192
      accepts_nested_attributes_for :diagnostic_measurements, :allow_destroy => true, reject_if: :all_blank

      has_many :evaluable_target_agreements,  -> { where(evaluate: true) },
               class_name: '::Portfolio::ItemTarget',
               foreign_key: :portfolio_item_id,
               inverse_of: nil

      has_many :other_target_agreements,  -> { where(evaluate: false) },
               class_name: '::Portfolio::ItemTarget',
               foreign_key: :portfolio_item_id,
               inverse_of: nil

      has_many :target_agreements,
               class_name: '::Portfolio::ItemTarget',
               foreign_key: :portfolio_item_id,
               inverse_of: :portfolio_item,
               dependent: :destroy # refs #192
      #accepts_nested_attributes_for :target_agreements, :allow_destroy => true, reject_if: :all_blank
      accepts_nested_attributes_for :target_agreements, :allow_destroy => true, reject_if: Proc.new {|p| p['content'].blank?}

      has_many :questionnaires,
               class_name: '::Questionnaire',
               foreign_key: :portfolio_item_id,
               inverse_of: :portfolio_item,
               dependent: :destroy
      accepts_nested_attributes_for :questionnaires, :allow_destroy => true, reject_if: :all_blank
      validates_associated :questionnaires

      has_many :checklists,
               class_name: '::Checklist',
               foreign_key: :portfolio_item_id,
               inverse_of: :portfolio_item,
               dependent: :destroy
      accepts_nested_attributes_for :checklists, :allow_destroy => true, reject_if: :all_blank

      has_and_belongs_to_many :checklist_questions, foreign_key: :portfolio_item_id, autosave: false # there is constraint in db

      has_many :portfolio_comments,
                class_name: '::Portfolio::Comment',
                foreign_key: :portfolio_item_id,
                inverse_of: :portfolio_item,
                dependent: :destroy

      def has_blocking_comment?
        portfolio_comments.latest.with_blocking_status.first.try(:item_blocked?)
      end

      scope :most_recently_closed,  -> { where(closed: true).order(start: :desc, updated_at: :desc) }

      validates :portfolio_step_id, :start, :portfolio_specialization_id, :level, presence: true
      validate do
        portfolio_step.item_requirements_validator.validate(self) unless portfolio_step.blank?
      end
      validates :level, inclusion: [0,1,2,3]
      validates :start,:parent_teacher_conference_date, :management_conference_date, date: true
      validate do
        if !closed?
          if self.class.where(closed: true,portfolio_id: portfolio_id).where('start > ?',start).any? ||
             ( self.new_record? &&
             self.class.where(closed: true,portfolio_id: portfolio_id,start: start).where('created_at > ?',created_at).any? )

            errors[:start] << "Ein Eintrag kann nicht zwischen abgeschlossenen Einträgen eingeschoben werden."
          end
        end
      end
      validates :parent_teacher_conference_date, chronological_order: {fields: [:start,:parent_teacher_conference_date]}
      validates :management_conference_date, chronological_order: {fields: [:start,:management_conference_date]}
      validates :content, length: { minimum: 10}, if: ::Proc.new { |record| record.content_is_attached == false }
      before_validation do
        # when the checkbox to include the author is checked, add the author to involved teachers
        # ----- the seperated checkbox is not used any more. refs #450
        # if author.class.name == 'Teacher'
        #   if creator_involved? # checkbox is checked
        #     involved_teachers << author unless involved_teachers.include?(author)
        #   else
        #     involved_teachers.delete(author)
        #   end
        # end
        # if author.class.name == 'Supporter'
        #   if creator_involved? # checkbox is checked
        #     involved_supporters << author unless involved_supporters.include?(author)
        #   else
        #     involved_supporters.delete(author)
        #   end
        # end
        if portfolio_step.present? &&  portfolio_step.subject_is_always_involved?
          self.subject_involved = true
        end
        self.level = portfolio_subject.portfolio_level if self.new_record?
        self.parent_teacher_conference = parent_teacher_conference_date.present?
        self.management_conference = management_conference_date.present?
        true
      end



      def diagnostic_ranges
        r = []
        r << 'Ökologie' if is_range_okologie?
        r << 'Integration' if is_range_integration?
        r << 'Intensität' if is_range_intensitaet?
        r
      end


      def collection_select_text
        "#{portfolio_step.name} ( #{portfolio_specialization.abbreviation} ) vom #{I18n.l(start,format: :short)}"
      end

      def level_description
        [ "0: Ausgangslage","1: Informelle Förderplanung", "2: Formelle Förderplanung", "3: Sonderpädagogische Förderung"][level]
      end



      # get if either any target agreements exist or if the step requires any
      def any_target_agreements?
        target_agreements.any? || portfolio_step.item_requirements_validator.has?('target_agreements')
      end
      alias :any_target_agreements :any_target_agreements?

      def any_dbrcs?
        portfolio_dbrcs.any?
      end
      alias :any_dbrcs :any_dbrcs?

      # get if either any measurements exist or if the step requires any
      #
      # In this case, measurements are also diagnostics and advancement_plans
      #
      def any_measurements?
        diagnostics.any? || advancement_plans.any? || measurements.any?  || class_diagnostics.any? || diagnostic_measurements.any?
            portfolio_step.item_requirements_validator.has?('diagnostic') ||
            portfolio_step.item_requirements_validator.has?('advancement_plan') ||
            portfolio_step.item_requirements_validator.has?('measurement') ||
            portfolio_step.item_requirements_validator.has?('diagnostic_measurement') ||
            portfolio_step.item_requirements_validator.has?('class_diagnostics')
      end
      alias :any_measurements :any_measurements?

      # get if there have been any results inserted for the step
      def any_results
        portfolio_step_result.present?
      end

      def institution_document_suggestions
        ids = []
        if portfolio_step.present?
          self.portfolio_subject.institution.institution_documents.where(file: nil).where(step_type: portfolio_step.type).each do |d|
            ids << d.all_children.where.not(file: nil).pluck(:id)
          end
        end
        InstitutionDocument.where(id: ids)
      end

      # get a description why this item cannot be discarded. This is useful when can_be_discarded? returned false
      #
      # as discarding involves closing of an item, the reasons for cannot be closed apply as well.
      #
      # @return [String] reason why cannot be discarded
      def cannot_be_discarded_reason
        cannot_be_closed_reason
      end

      # get if the item can be discarded
      #
      # an item can be discarded when it can be closed
      #
      # @return [boolean]
      def can_be_discarded?
        can_be_closed?
      end

      # discards the item.
      #
      # it is closed then, but the discarded flag is already set,
      # so the item is not included in portfolio rule validations.
      def discard
        update_attribute(:discarded, true)
        update_attribute(:closed, true)
      end


      # get a description why this item cannot be closed. This is useful when can_be_closed? returned false
      #
      # @return [String] reason why cannot be closed
      def cannot_be_closed_reason
        unless  questionnaires.inject(true){|momo,q| momo && q.is_completed? }
          return "Der Eintrag hat noch offene Fragebögen. Fragebögen müssen komplett ausgefüllt sein und ein Beurteiler muss eingetragen sein"
        end
        if has_open_diagnostics?
          return "Der Eintrag hat noch offene Diagnostiken"
        end
        if closed?
          return "Eintrag ist schon abgeschlossen"
        end
        if (start > ::Date.today)
          return "Der Eintrag hat noch nicht begonnen."
        end
        if ( needs_evaluation && !was_evaluated_completely )
          return "Der Eintrag hat noch offene Evaluationen"
        end
        if ( self.class.where(closed: false,portfolio_id: portfolio_id).where('start < ?',start).any? ||
            self.class.where(closed: false,portfolio_id: portfolio_id,start: start).where('updated_at < ?',updated_at).any?)
          return "Der Eintrag kann erst abgeschlossen werden, wenn alle früheren Einträge abgeschlossen sind"
        end
      end

      def has_open_diagnostics?
        diagnostics.exists?(results_closed: false) ||
            class_diagnostics.exists?(results_closed: false) ||
            diagnostic_measurements.exists?(results_closed: false) ||
            questionnaires.inject(false){|momo,q| momo || !q.is_completed?}
      end

      def has_any_diagnostics?
        diagnostics.any? ||
            class_diagnostics.any? ||
            diagnostic_measurements.any? ||
            questionnaires.any?
      end



      # get if the item can be closed
      #
      # this checks the following
      # - item is not already closed
      # - item has started
      # - all evaluations are done
      # - there is no unclosed item chronologiccally before a closed item.
      #
      # @return [String] reason why cannot be closed
      def can_be_closed?
        # all diagnostic results must be closed
        return false unless questionnaires.inject(true){|momo,q|momo && q.is_completed? }
        return false if has_open_diagnostics?
        return false if closed? # must be open to be closed
        return false if start > ::Date.today # must not be in the future to be closed
        return false if needs_evaluation && !was_evaluated_completely # must have no open evaluation to be closed
        return false if self.class.where(closed: false,portfolio_id: portfolio_id).where('start < ?',start).any? # there should be no gaps
        return false if self.class.where(closed: false,portfolio_id: portfolio_id,start: start).where('updated_at < ?',updated_at).any? # there should be no gaps on same date
        return true
      end

      # close this item.
      #
      # can_be_closed? must be checked before!
      #
      # Closed items can only be changed with some restriction, this is to ensure that a once valid portfolio, validated
      # successfully against the rule chain, becomes invalid, because an old item is edited, so that it does not
      # fulfill the rule requirements any more.
      # This could also be prevented by validation, but it is rather difficult to communicate the exact reason, why the change
      # is not possible to the user, so this is left for later versions.
      # @return [Boolean] true when the item is really closed
      def close
        update_attribute(:closed, true)
        update_attribute(:discarded, false)
      end

      # check if an item can be deleted
      #
      # @return [Boolean] true when the item can be deleted
      def can_be_deleted?
        !closed
      end

      # get the reason why the item cannot be deleted
      #
      # @return [Boolean] true when the item can be deleted
      def cannot_be_deleted_reason
        "Eintrag ist schon abgeschlossen"
      end

      # get the records that are allowed to be set as involved teachers for this portfolio
      # which are teachers at the same institution as the portfolio ( author is excluded, as it is set via checkbox)
      #
      # this is used for the select involved teachers field in portfolio item create/update form
      # @return [ActiveRecord::Associations::CollectionProxy<::Teacher>]
      def available_involved_teachers
        portfolio_subject.portfolio_item_available_involved_teachers(self)
      end

      # get the records that are  allowed to be set as involved guardians for this portfolio
      #
      # for base items, guardians of clients at the same institution as the portfolio are returned
      # this is used for the select involved guardians field in portfolio item create/update form
      #
      # @return [ActiveRecord::Associations::CollectionProxy<::Guardian>]
      def available_involved_guardians
        portfolio_subject.portfolio_item_available_involved_guardians(self)
      end


      # get the records that are allowed to be set as involved supporters
      # which are supporters related all clients in the same institution as the portfolio
      #
      # this is used by the select involved supporters field in portfolio item create/update form
      # @return [ActiveRecord::Associations::CollectionProxy<::Supporter>]
      def available_involved_supporters
        portfolio_subject.portfolio_item_available_involved_supporters(self)
      end

      # get all portfolio specializations that can be set for the item
      #
      # this is used by the select portfolio_specialization field in portfolio item create/update form
      # @return [ActiveRecord::Associations::CollectionProxy<::PortfolioSpecialization>]
      def available_specializations
        portfolio_subject.available_specializations(self)
      end

      # can this item have teacher involvements?
      #
      # this is used in portfolio item create/update form
      def allows_involved_teachers?
        true # currently, all items allow teacher
      end

      # can this item have supporter involvements?
      #
      # this is used in portfolio item create/update form
      def allows_involved_supporters?
        true # currently, all items allow supporter
      end

      # can this item have guardian involvements?
      #
      # this is used in portfolio item create/update form
      def allows_involved_guardians?
        portfolio_subject.allows_guardians?(self)
      end

      # the item needs an evaluation
      #
      # this is true, when any target agreements with evaluations are present.
      # used for display of form tabs
      #
      # @return [Boolean]
      def needs_evaluation
        # wurden evaluierbare Zielvereinbarungen getroffen?
        return (evaluable_target_agreements.size > 0)
      end


      # the item needs an evaluation and not all evaluations are completed
      #
      # @return [Boolean]
      def was_evaluated_partially
        return false unless needs_evaluation # es wird keine Evaluierung benötigt
        evaluable_target_agreements.each do |ta|
          return true if ta.evaluation_completed
        end
        false
      end

      # return all documents that belong to informal results of this item
      def result_documents
        docs = []
        diagnostics.each do |d|
          d.informal_results.each do |i|
            i.documents.each do |doc|
              docs << doc
            end
          end
        end
        diagnostic_measurements.each do |d|
          d.informal_results.each do |i|
            i.documents.each do |doc|
              docs << doc
            end
          end
        end
        class_diagnostics.each do |d|
          d.informal_results.each do |i|
            i.documents.each do |doc|
              docs << doc
            end
          end
        end
        docs
      end

      # the item needs an evaluation and all evaluations are completed
      #
      # @return [Boolean]
      def was_evaluated_completely
        return false unless needs_evaluation # es wird keine Evaluierung benötigt
        evaluable_target_agreements.each do |ta|
          return false unless ta.evaluation_completed
        end
        true
      end

      # get the current portfolio level
      #
      # @return [Boolean]
      def portfolio_level
        return portfolio_subject.portfolio_level if new_record?
        return level
      end

    end
end