class Questionnaire < ActiveRecord::Base
  def self.available_types
    [::Questionnaires::SdqTeacher,::Questionnaires::SdqGuardian,::Questionnaires::SdqClient,
    ::Questionnaires::ItrfTeacher]
  end

  belongs_to :supporter
  belongs_to :teacher
  belongs_to :client
  belongs_to :guardian
  belongs_to :portfolio_item, class_name: '::Portfolio::Item'
  has_many :questionnaire_results,
           class_name: '::QuestionnaireResult',
           foreign_key: :questionnaire_id,
           inverse_of: :questionnaire,
           dependent: :destroy
  accepts_nested_attributes_for :questionnaire_results, :allow_destroy => false
  has_many :portfolio_dbrcs,
           class_name: '::Portfolio::Dbrc',
           foreign_key: :questionnaire_id,
           inverse_of: :questionnaire,
           dependent: :destroy
  validates_presence_of  :portfolio_item

  validate do
    unless has_any_persons? || new_record?
      errors.add(:teacher_id,"Ein Lehrer oder eine andere Person muss angegeben sein.") if accepts_teacher?
      errors.add(:client_id,"Ein Schüler oder eine andere Person muss angegeben sein.") if accepts_client?
      errors.add(:supporter_id,"Ein Externer oder eine andere Person muss angegeben sein.") if accepts_supporter?
      errors.add(:guardian_id,"Ein Erziehungsberechtigter oder eine andere Person muss angegeben sein.") if accepts_guardian?
    end
  end

  before_save do
    if supporter
      self.supporter_name=supporter.full_name
    end
    if teacher
      self.teacher_name=teacher.full_name
    end
    if client
      self.client_name=client.full_name
    end
    if guardian
      self.guardian_name=guardian.full_name
    end
  end

  before_create do
    questions.each do |question|
      questionnaire_results.build(question_id: question.id, group_id: question.group_id, question_order: question.order)
    end
  end

  def is_completed?
    questionnaire_results.where(result: nil).count == 0 && has_any_persons?
  end

  def accepts_teacher?
    false
  end

  def accepts_client?
    false
  end

  def accepts_guardian?
    false
  end

  def accepts_supporter?
    true
  end

  def accepts_person?
    true
  end

  def assessor_names
    [teacher_name,guardian_name,supporter_name,client_name,person_name].reject(&:blank?).inject([]){|m,p| m << p }
  end


  def has_any_persons?
    [teacher,guardian,supporter,client,person_name].inject(false){|m,p| m || p.present?}
  end

  # get the records that are allowed to be set as assessing teachers
  def available_involved_teachers
    portfolio_item.try(:available_involved_teachers).to_a # portfolio_item is probably nil
  end

  # get the records that are  allowed to be set as assessing guardians
  def available_involved_guardians
     portfolio_item.try(:available_involved_guardians).to_a # portfolio_item is probably nil
  end

  # get the records that are allowed to be set as assessing supporters
  def available_involved_supporters
    portfolio_item.try(:available_involved_supporters).to_a # portfolio_item is probably nil
  end

  # get the records that are allowed to be set as assessing clients
  def available_involved_clients
    if portfolio_item.portfolio.subject.is_a? Client
      return [portfolio_item.portfolio.subject]
    end
    if portfolio_item.portfolio.subject.is_a? ClientGroup
      return portfolio_item.portfolio.subject.clients
    end
  end

  def self.questionnaire_name
    "Fragebogen"
  end

  def questionnaire_name
    self.class.questionnaire_name
  end

  # @return Array[String]
  def documents
    []
  end

  def link
    "http://www.example.com"
  end

  # @return Array[QuestionGroup]
  def groups
    []
  end

  # @return Array[Analysis]
  def analyses
    []
  end

  # @return Array[Question]
  def questions
    groups.map(&:questions).flatten
  end

  # Result tests and
  # @return Array[String]
  def result_texts
    []
  end

  class Question
    attr_accessor :id,:order, :name, :values, :group_id, :drc_items
    def initialize(id,order,name,values,drc_items=[])
      self.id=id
      self.order=order
      self.name=name
      self.values=values
      self.drc_items=drc_items
    end
  end

  class DrcItem
    attr_accessor :id,:order, :name
    def initialize(id,order,name)
      self.id=id
      self.order=order
      self.name=name
    end
  end

  class QuestionGroup
    attr_accessor :id,:order, :name, :questions
    def initialize(id,order,name,questions)
      self.id=id
      self.order=order
      self.name=name
      self.questions=questions.map do |question|
        question.group_id=id
        question
      end
    end
  end

  class Analysis
    attr_accessor :id, :order, :name, :value, :text, :color, :interpolated
    def initialize(id,order,name,value, options={})
      self.id=id
      self.order=order
      self.name=name
      self.value=value
      self.interpolated=(value.to_i!=value)
    end
  end

  class TernaryProblemAnalysis < Analysis
    def initialize(id,order,name,value, options={})
      super
      if value < options[:left_maximum]
        self.color='success'
        self.text="Normal"
      end
      if value >= options[:left_maximum] && value < options[:right_minimum]
        self.color='warning'
        self.text="Grenzwertig"
      end
      if value >= options[:right_minimum]
        self.color='danger'
        self.text="Auffällig"
      end
    end
  end

  class TernaryStrengthAnalysis < Analysis
    def initialize(id,order,name,value, options={})
      super
      if value < options[:left_maximum]
        self.color='danger'
        self.text="Auffällig"
      end
      if value >= options[:left_maximum] && value < options[:right_minimum]
        self.color='warning'
        self.text="Grenzwertig"
      end
      if value >= options[:right_minimum]
        self.color='success'
        self.text="Normal"
      end
    end
  end


end
